
/* 
** AUTHORS: 
** Rama Alebouyeh (rama[at]cs.ucsb.edu)
** Matthew S Allen (msa[at]cs.ucsb.edu)
*/

#include <stdio.h>
#include <string.h>
#include <pthread.h>
#include <stdlib.h>
#include "route.h"
#include "host.h"
#include "key.h"
#include "jrb.h"
#include "jval.h"
#include "log.h"


#define LEAFSET_SIZE 8   /* (must be even) excluding node itself */
#define MAX_ROW KEY_SIZE/BASE_B
#define MAX_COL power(2,BASE_B)

struct {
  ChimeraHost *me;
  char *keystr;
  ChimeraHost ***table;
  ChimeraHost **leftleafset;
  ChimeraHost **rightleafset;
  pthread_mutex_t lock;
} ROUTE_GLOBAL;

void leafset_update(ChimeraHost *host, int joined, ChimeraHost **deleted, ChimeraHost **added);
int leafset_size(ChimeraHost **arr);
void leafset_insert(ChimeraHost *host ,int right_or_left, ChimeraHost **deleted, ChimeraHost **added); 
void leafset_delete(ChimeraHost *host, int right_or_left, ChimeraHost **deleted);
void leafset_print();
void sort_hosts (ChimeraHost **hosts,Key key,int size);
int power ( int base, int n);
int hexalpha_to_int(int c);

void route_keyupdate(ChimeraHost *me) {
  ROUTE_GLOBAL.me = me;
  sprintf(ROUTE_GLOBAL.keystr,"%s", ROUTE_GLOBAL.me->key.keystr);
}

void route_init(ChimeraHost * me) {

  int i,j;

  /* allocate memory for routing table */
  ROUTE_GLOBAL.table=(ChimeraHost ***)malloc(sizeof (struct ChimeraHost **) * MAX_ROW);
  for ( i=0; i< MAX_ROW; i++){
     ROUTE_GLOBAL.table[i]=(ChimeraHost **)malloc(sizeof(ChimeraHost *)*MAX_COL);
     for(j=0;j<MAX_COL;j++)
       ROUTE_GLOBAL.table[i][j]=NULL;
     
  }
 
  ROUTE_GLOBAL.keystr = (char *) malloc(sizeof(char)*MAX_ROW);
  route_keyupdate(me);
  
  
  /* allocate memory for leafsets */
  ROUTE_GLOBAL.leftleafset=(ChimeraHost **)malloc(sizeof(ChimeraHost *)*(LEAFSET_SIZE/2)+1);
  ROUTE_GLOBAL.rightleafset=(ChimeraHost **)malloc(sizeof(ChimeraHost *)*(LEAFSET_SIZE/2)+1);
  for(i=0;i<(LEAFSET_SIZE/2)+1;i++){
    ROUTE_GLOBAL.leftleafset[i]=NULL;
    ROUTE_GLOBAL.rightleafset[i]=NULL;
  }
  
  pthread_mutex_init(&ROUTE_GLOBAL.lock, NULL);

}

ChimeraHost **route_get_table(void){
  ChimeraHost **ret;
  int i,j,k,count;

  pthread_mutex_lock(&ROUTE_GLOBAL.lock);
  
  count = 0;
  
  for(i=0;i<MAX_ROW;i++)
    for(j=0;j<MAX_COL;j++)
      if(ROUTE_GLOBAL.table[i][j]!=NULL)
	count++;
  ret = (ChimeraHost **) malloc(sizeof(ChimeraHost *) * (count + 1));
  k=0;
  for(i=0;i<MAX_ROW;i++)
    for(j=0;j<MAX_COL;j++)
      if(ROUTE_GLOBAL.table[i][j]!=NULL)
	ret[k++]= host_get(ROUTE_GLOBAL.table[i][j]->name, ROUTE_GLOBAL.table[i][j]->port);

  ret[k]=NULL;
  pthread_mutex_unlock(&ROUTE_GLOBAL.lock);
  return ret;

}

ChimeraHost **route_row_lookup(Key key){
  ChimeraHost **ret;
  int i,j,k,count;

  pthread_mutex_lock(&ROUTE_GLOBAL.lock);

  
  i= key_index(ROUTE_GLOBAL.me->key,key);

  /* find out the number of hosts exists in the matching row */
  count = 0;
  for(j=0;j<MAX_COL;j++)
    if(ROUTE_GLOBAL.table[i][j]!=NULL)
      count++;

  ret = (ChimeraHost **) malloc(sizeof(ChimeraHost *) * (count + 2));
  k=0;
  for(j=0;j<MAX_COL;j++)
    if(ROUTE_GLOBAL.table[i][j]!=NULL)
      ret[k++]= host_get(ROUTE_GLOBAL.table[i][j]->name, ROUTE_GLOBAL.table[i][j]->port);

  ret[k++]=host_get(ROUTE_GLOBAL.me->name,ROUTE_GLOBAL.me->port);
  ret[k]=NULL;

  pthread_mutex_unlock(&ROUTE_GLOBAL.lock);

  return ret;
}


ChimeraHost **route_lookup(Key key, int count, int is_safe){
  int i,j;
  ChimeraHost *leaf, *tmp;
  int Lsize,Rsize;
  Key Lrange,Rrange,dif1,dif2;
  ChimeraHost **ret;
  Key diffs[LEAFSET_SIZE+1+(MAX_ROW * MAX_COL)];
  int index=0;
  int size = 0;
  ChimeraHost **hosts;
  
  pthread_mutex_lock(&ROUTE_GLOBAL.lock);
  
  log_message(LOG_ROUTING, "%s: lookup for key %s!\n",ROUTE_GLOBAL.me->key.keystr,key.keystr);
  

  /* brute force method to solve count requirements */

  /*calculate the leafset and table size*/
  Lsize=leafset_size(ROUTE_GLOBAL.leftleafset);
  Rsize=leafset_size(ROUTE_GLOBAL.rightleafset);


  /* confirm the leafset range */
  if(Rsize==0)
    key_assign(&Rrange, ROUTE_GLOBAL.me->key);
  else
    key_assign(&Rrange, ROUTE_GLOBAL.rightleafset[Rsize-1]->key);

  if(Lsize==0)
    key_assign(&Lrange, ROUTE_GLOBAL.me->key);
  else   
    key_assign(&Lrange, ROUTE_GLOBAL.leftleafset[Lsize-1]->key);
    
  hosts =  (ChimeraHost **) malloc(sizeof(ChimeraHost *) * (LEAFSET_SIZE+1+(MAX_ROW * MAX_COL)));

  
  leaf = ROUTE_GLOBAL.me;
  log_message(LOG_ROUTING, " me: (%s, %d, %s)\n", leaf->name, leaf->port, leaf->key.keystr);
  
  hosts[index++]= ROUTE_GLOBAL.me;
 
  
  /* look left */
  for(i = 0; i < Lsize; i++) {
    leaf = ROUTE_GLOBAL.leftleafset[i];
    log_message(LOG_ROUTING, " Left[%d] (%s, %d, %s): ",i, leaf->name, leaf->port, leaf->key.keystr);
    hosts[index++] = leaf;
  }
  
  /* look right */
  for(i = 0; i < Rsize; i++) {
    leaf = ROUTE_GLOBAL.rightleafset[i];
    log_message(LOG_ROUTING, " Right[%d] (%s, %d, %s)\n",i, leaf->name, leaf->port, leaf->key.keystr);
    hosts[index++] =leaf; 
  }
  
  i=key_index(ROUTE_GLOBAL.me->key,key);
    
  for(j=0;j<MAX_COL;j++)
    if(ROUTE_GLOBAL.table[i][j]!=NULL) {
      leaf = ROUTE_GLOBAL.table[i][j];
      log_message(LOG_ROUTING, " Table[%d][%d] (%s, %d, %s)\n",i,j,leaf->name, leaf->port, leaf->key.keystr);
      hosts[index++] =leaf; 
    }
  hosts[index]=NULL;
  sort_hosts (hosts, key,index);
  
  /* find the best #count# entries that we looked at... could be much better */
  ret = (ChimeraHost **) malloc(sizeof(ChimeraHost *) * (count + 1));
  for(i = 0, j = 0; (i - j) < count && hosts[i]!=NULL; i++) {
    tmp = hosts[i];
    
    if((i - j) > 0 && key_equal(ret[(i-j)-1]->key ,tmp->key)) {
      j++;
      continue;
    }
    log_message(LOG_ROUTING, " Result[%d] (%s, %d, %s)\n",i, tmp->name, tmp->port, tmp->key.keystr);
    ret[i-j] = host_get(tmp->name, tmp->port);
  }
  ret[i-j] = NULL;

  // to prevent bouncig
  key_distance(&dif1,key,ret[0]->key);
  key_distance(&dif2,key,ROUTE_GLOBAL.me->key);  
  if(count ==1 && key_equal(dif1, dif2))
    ret[0]=ROUTE_GLOBAL.me;
  
    
  pthread_mutex_unlock(&ROUTE_GLOBAL.lock);

  
  if(ret[0]!=NULL)
    log_message(LOG_ROUTING, "RESULT: %s\n",ret[0]->key.keystr);
  else 
    log_message(LOG_ROUTING, "RESULT: %s\n",ROUTE_GLOBAL.me->key.keystr);

  return(ret);
}

void sort_hosts (ChimeraHost **hosts,Key key,int size){
  int  i,j;
  ChimeraHost *tmp;
  Key dif1;
  Key dif2;
  
  for(i=0;i<size;i++){
    for(j=i+1;j<size;j++){
      if(hosts[i]!=NULL && hosts[j]!=NULL){
	key_distance(&dif1,hosts[i]->key,key);
	key_distance(&dif2,hosts[j]->key,key);	
	if(key_comp(dif2,dif1)<0){		
	  tmp = hosts[i];
	  hosts[i]=hosts[j];
	  hosts[j]=tmp;
	}
      }
      
    }
  }
  
}

int leafset_size(ChimeraHost **arr){
  int i;
  for ( i=0;arr[i]!=NULL;i++);
  return i;

 
}

void leafset_update(ChimeraHost *host, int joined, ChimeraHost **deleted, ChimeraHost **added){

  int Lsize=0;
  int Rsize=0;
  Key midpoint;
  
  Lsize=leafset_size(ROUTE_GLOBAL.leftleafset);
  Rsize=leafset_size(ROUTE_GLOBAL.rightleafset);

  key_midpoint(&midpoint,ROUTE_GLOBAL.me->key);
  
  if(joined)
    {
      /* key falls in the right side of node */
      if( key_between (host->key,ROUTE_GLOBAL.me->key, midpoint)){
	if(Rsize<LEAFSET_SIZE/2 || key_between( host->key,ROUTE_GLOBAL.me->key,ROUTE_GLOBAL.rightleafset[Rsize-1]->key)){
	  /* insert in Right leafset */
	  leafset_insert(host,1,deleted,added); 
	}
      }
      
      /* key falls in the left side of the node*/
      else 
	if(Lsize<LEAFSET_SIZE/2 || key_between( host->key,ROUTE_GLOBAL.leftleafset[Lsize-1]->key,ROUTE_GLOBAL.me->key ))
	  /* insert in Left leafset */
	  leafset_insert(host,0,deleted,added);
    }
  else
    {
      if( key_between (host->key,ROUTE_GLOBAL.me->key, midpoint)){
	leafset_delete(host,1,deleted);
      }
      else 
	leafset_delete(host,0,deleted);
    }

}

extern void chimera_update_upcall(Key *k, ChimeraHost *h, int joined);

void leafset_delete(ChimeraHost *host, int right_or_left, ChimeraHost **deleted){
  int i=0 ,size;
  int match=0;
  ChimeraHost **p;
  
  /*inert in right leafset*/
  if(right_or_left==1){
    size=leafset_size(ROUTE_GLOBAL.rightleafset);
    p=ROUTE_GLOBAL.rightleafset;
  }
  /*insert in left leafset*/
  else{
    size=leafset_size(ROUTE_GLOBAL.leftleafset);
    p=ROUTE_GLOBAL.leftleafset;
  }


  for(i=0;i<size && !(key_equal(p[i]->key,host->key)) ;i++);
  if(i<size){
    *deleted = p[i];
    match=1;
  }

  if(match){
    do{
      p[i]=p[i+1];
      i++;
    } while(i<size-1);
    p[i]=NULL;
  }


}

void leafset_insert(ChimeraHost *host, int right_or_left, ChimeraHost **deleted, ChimeraHost **added){

  int i=0 ,size;
  ChimeraHost **p;
  ChimeraHost *tmp1,*tmp2;  
  ChimeraHost *input = host;
  Key dif1,dif2;

  /*inert in right leafset*/
  if(right_or_left==1){
    size=leafset_size(ROUTE_GLOBAL.rightleafset);
    p=ROUTE_GLOBAL.rightleafset;
  }
  /*insert in left leafset*/
  else{
    size=leafset_size(ROUTE_GLOBAL.leftleafset);
    p=ROUTE_GLOBAL.leftleafset;
  }
   
  if(size==0) {
    p[0]=input;
    *added = input;
  }

  else {
    key_distance(&dif1,ROUTE_GLOBAL.me->key, p[i]->key);
    key_distance(&dif2,ROUTE_GLOBAL.me->key, host->key);
    while((i<size) && (key_comp(dif1,dif2)<=0)){
      if(key_equal(p[i]->key,input->key)){
	 host_release(input);
	 return;
      }
      i++;
      if(i<size){
	key_distance(&dif1,ROUTE_GLOBAL.me->key, p[i]->key);
	key_distance(&dif2,ROUTE_GLOBAL.me->key, host->key);
      }
    }
    
    tmp1=input;
    *added = input;
    
    while(i<LEAFSET_SIZE/2){
      tmp2=p[i];
      p[i++]=tmp1;
      tmp1=tmp2;
    } 

    /* there is leftover */
    if(size==LEAFSET_SIZE/2 && tmp2 != NULL)
      *deleted = tmp2;
  }
}

ChimeraHost **route_neighbors(int count){
  int i=0,j=0;
  int Rsize=0;
  int Lsize=0;
  int ret_size;
  ChimeraHost **ret=(ChimeraHost **)malloc(sizeof(ChimeraHost *)*(count+1));
  ChimeraHost *tmp;
  int index=0;
  ChimeraHost **hosts=(ChimeraHost **)malloc(sizeof(ChimeraHost *)*(LEAFSET_SIZE+1));

  pthread_mutex_lock(&ROUTE_GLOBAL.lock);

  Lsize=leafset_size(ROUTE_GLOBAL.leftleafset);
  Rsize=leafset_size(ROUTE_GLOBAL.rightleafset);

  if(count>Rsize+Lsize){
    ret_size=Rsize+Lsize;
  }
  else 
    ret_size=count;

  /* creat a jrb of leafset pointers sorted on distance */
  for(i=0;i<Lsize;i++) {
    tmp = ROUTE_GLOBAL.leftleafset[i];
    hosts[index++]=tmp;
  }
  
  for(i=0;i<Rsize;i++) {
    tmp = ROUTE_GLOBAL.rightleafset[i];
    hosts[index++]=tmp;
  }

  hosts[index]=NULL;
  /* sort aux */
  sort_hosts (hosts,ROUTE_GLOBAL.me->key,index);
  
  for(i=0;i<ret_size;i++) {
    tmp = hosts[i];
    ret[i] = host_get(tmp->name, tmp->port);
  }
  
  ret[i] = NULL;

  pthread_mutex_unlock(&ROUTE_GLOBAL.lock);
    
  return ret;
}



void route_update(ChimeraHost *host, int joined){
  int i,j;
  ChimeraHost *deleted = NULL, *added = NULL;

  pthread_mutex_lock(&ROUTE_GLOBAL.lock);

  
  if(key_equal(ROUTE_GLOBAL.me->key, host->key)) {
    pthread_mutex_unlock(&ROUTE_GLOBAL.lock);
    return;
  }

  i = key_index(ROUTE_GLOBAL.me->key, host->key);
  j=hexalpha_to_int(host->key.keystr[i]);

  /*join*/
  if(joined) 
    {
      if(ROUTE_GLOBAL.table[i][j]!=NULL)
	host_release(ROUTE_GLOBAL.table[i][j]);
      
      ROUTE_GLOBAL.table[i][j]=host_get(host->name, host->port);
      leafset_update(host,joined,&deleted,&added);
    }
  
  /*delete*/
  else 
    {
      if(ROUTE_GLOBAL.table[i][j]!=NULL && key_equal(ROUTE_GLOBAL.table[i][j]->key,host->key))
	{
	  host_release(ROUTE_GLOBAL.table[i][j]);
	  ROUTE_GLOBAL.table[i][j]=NULL;
	}

      leafset_update(host,joined,&deleted,&added);
	
    }

  pthread_mutex_unlock(&ROUTE_GLOBAL.lock);

  if(deleted != NULL) {
    chimera_update_upcall(&(deleted->key),deleted,0);
  }
  if(added != NULL) {
    chimera_update_upcall(&(added->key),added,1);
  }
}

void gethostinfo(ChimeraHost *CHost){

  char tmp[64];
  CHost->name=(char *)malloc(sizeof(char)*64);
  printf("Enter key: ");
  scanf("%s", tmp);
  str_to_key(tmp, &(CHost->key));
  printf("Key :");
  key_print(CHost->key);
}

int hexalpha_to_int(int c)
{
  char hexalpha[] = "0123456789abcdef";
  int i;
  int answer = 0;

  for(i = 0; answer == 0 && hexalpha[i] != '\0'; i++)
  {
    if(hexalpha[i] == c)
    {
      answer = i ;
    }
  }

  return answer;
}

int power ( int base, int n){
  int i,p=1;

  for(i=1; i<=n; i++)
    p=p*base;
  return p;
}

void leafset_print(){
  int i;
  int Lsize,Rsize;
  Lsize=leafset_size(ROUTE_GLOBAL.leftleafset);
  Rsize=leafset_size(ROUTE_GLOBAL.rightleafset);
  
  fprintf(stderr,"LEFT: ");
  for (i=0;i<Lsize;i++)
    fprintf(stderr,"%s ", ROUTE_GLOBAL.leftleafset[i]->key.keystr);
  fprintf(stderr,"\nRIGHT: ");
  for (i=0;i<Rsize;i++)
    fprintf(stderr,"%s ", ROUTE_GLOBAL.rightleafset[i]->key.keystr);
  fprintf(stderr,"\n");
}

void printTable(){

  int i,j;


 /* print the table */

  fprintf(stderr,"------------------------------- TABLE-------------------------------\n");
  for ( i=0; i< MAX_ROW; i++){
    for(j=0;j<MAX_COL;j++){
      if( ROUTE_GLOBAL.table[i][j]!=NULL)
	fprintf(stderr,"%s ", ROUTE_GLOBAL.table[i][j]->key.keystr);
      else
	fprintf(stderr,"00000000 00000000 00000000 00000000  00000000");
    }
    fprintf(stderr,"\n");
  }
  fprintf(stderr,"----------------------------------------------------------------------\n");
}



